--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Keep a running sum for each element in snk_in_arr[i]
-- Description:
-- . A RAM-block is used to store the running sums as wel as to output the
--   corresponding previous running sum so both align at the adder inputs.

ENTITY corr_accumulator IS
  GENERIC (
    g_nof_inputs         : NATURAL; -- Number of input streams
    g_nof_accumulators   : NATURAL; -- Number of accumulators to keep per input stream
    g_integration_period : NATURAL; -- Number of timesamples to accumulate
    g_data_w             : NATURAL  -- Complex input data width
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0)
  );
END corr_accumulator;

ARCHITECTURE str OF corr_accumulator IS

  -- Complex accumulator data width: wide enough to support g_integration_period accumulations
--  CONSTANT c_acc_data_w               : NATURAL := ceil_log2(g_integration_period*(pow2(g_data_w)-1));
  CONSTANT c_acc_data_w               : NATURAL := g_data_w + ceil_log2(g_integration_period);

  -- c_shiftram_delay is such that common_shiftram output aligns exactly with snk_in_arr. Functionally this 
  -- means we aligned the current word to the corresponding previous word at the adder inputs.
  CONSTANT c_shiftram_io_delay        : NATURAL := 4; -- common_shiftram data_in takes 4 cycles to emerge as data_out

  -- Note: Increase corr_adder pipelining by n = decrease c_shiftram_io_delay by n and vice versa.
  CONSTANT c_shiftram_delay           : NATURAL := g_nof_accumulators-c_shiftram_io_delay-2;

  CONSTANT c_shift_w                  : NATURAL := ceil_log2(g_nof_accumulators);
  CONSTANT c_common_shiftram_shift_in : STD_LOGIC_VECTOR(c_shift_w-1 DOWNTO 0) := TO_UVEC(c_shiftram_delay, c_shift_w);

  -- Counter to keep track of the total numer of accumulations so we know when to reset accumulator to zero
  SIGNAL acc_cnt                      : STD_LOGIC_VECTOR(ceil_log2(g_nof_accumulators*g_integration_period)-1 DOWNTO 0);
  SIGNAL nxt_acc_cnt                  : STD_LOGIC_VECTOR(ceil_log2(g_nof_accumulators*g_integration_period)-1 DOWNTO 0);

  SIGNAL corr_adder_snk_in_2arr_2     : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0); -- Array of pairs
  SIGNAL nxt_corr_adder_snk_in_2arr_2 : t_dp_sosi_2arr_2(g_nof_inputs-1 DOWNTO 0); -- Array of pairs
  SIGNAL corr_adder_src_out_arr       : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  SIGNAL common_shiftram_snk_in_arr   : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL common_shiftram_src_out_arr  : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

  TYPE t_common_shiftram_data_out_shift_arr IS ARRAY(g_nof_inputs-1 DOWNTO 0) OF STD_LOGIC_VECTOR(c_shift_w-1 DOWNTO 0);
  SIGNAL common_shiftram_data_out_shift_arr : t_common_shiftram_data_out_shift_arr;

  SIGNAL nxt_src_out_arr              : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- Adder inputs: current snk_in_arr + corresponding previous running sum
  -- from shiftram.
  -- When  we've completed one integration period for all the channels,
  -- we'll feed zeros into the second adder input to reset the accumulator value
  -- for each channel.
  -----------------------------------------------------------------------------
  nxt_acc_cnt <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt)=g_nof_accumulators*g_integration_period-1 ELSE
                 INCR_UVEC(acc_cnt, 1) WHEN snk_in_arr(0).valid = '1' ELSE acc_cnt;

  gen_adder_inputs : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    -- Adder input 0: multiplier output
    nxt_corr_adder_snk_in_2arr_2(i)(0).valid                   <= snk_in_arr(i).valid;
    nxt_corr_adder_snk_in_2arr_2(i)(0).re(g_data_w-1 DOWNTO 0) <= snk_in_arr(i).re(g_data_w-1 DOWNTO 0);
    nxt_corr_adder_snk_in_2arr_2(i)(0).im(g_data_w-1 DOWNTO 0) <= snk_in_arr(i).im(g_data_w-1 DOWNTO 0);
    -- . Pad the complex data width zeroes so all adder input bits are defined
    nxt_corr_adder_snk_in_2arr_2(i)(0).re(c_acc_data_w-1 DOWNTO g_data_w) <= (OTHERS=>'0');
    nxt_corr_adder_snk_in_2arr_2(i)(0).im(c_acc_data_w-1 DOWNTO g_data_w) <= (OTHERS=>'0');

    -- Adder input 1: shiftram output
    nxt_corr_adder_snk_in_2arr_2(i)(1).valid                       <= common_shiftram_src_out_arr(i).valid; 
    nxt_corr_adder_snk_in_2arr_2(i)(1).re(c_acc_data_w-1 DOWNTO 0) <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt)<=g_nof_accumulators-1 ELSE
                                                                      common_shiftram_src_out_arr(i).data(2*c_acc_data_w-1 DOWNTO c_acc_data_w);
    nxt_corr_adder_snk_in_2arr_2(i)(1).im(c_acc_data_w-1 DOWNTO 0) <= (OTHERS=>'0') WHEN TO_UINT(acc_cnt)<=g_nof_accumulators-1 ELSE
                                                                      common_shiftram_src_out_arr(i).data(  c_acc_data_w-1 DOWNTO 0);  
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Complex adder stage
  -----------------------------------------------------------------------------
  u_corr_adder : ENTITY work.corr_adder
  GENERIC MAP (
    g_nof_inputs => g_nof_inputs,
    g_data_w     => c_acc_data_w
  )
  PORT MAP (
    clk            => clk,
    rst            => rst,

    snk_in_2arr_2  => corr_adder_snk_in_2arr_2,
    src_out_arr    => corr_adder_src_out_arr
  );  

  -----------------------------------------------------------------------------
  -- Write the current sum to RAM; RAM outputs delayed running sums that align
  -- at the adder inputs:
  -- . common_shiftram_src_out_arr = delayed corr_adder_src_out_arr
  -----------------------------------------------------------------------------
  gen_concat_complex : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    -- Concatenate real&imaginary parts
    common_shiftram_snk_in_arr(i).data(2*c_acc_data_w-1 DOWNTO c_acc_data_w) <= corr_adder_src_out_arr(i).re(c_acc_data_w-1 DOWNTO 0);
    common_shiftram_snk_in_arr(i).data(  c_acc_data_w-1 DOWNTO 0)            <= corr_adder_src_out_arr(i).im(c_acc_data_w-1 DOWNTO 0);
    common_shiftram_snk_in_arr(i).valid                                      <= corr_adder_src_out_arr(i).valid;
  END GENERATE;

  gen_common_shiftram : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_common_shiftram : ENTITY common_lib.common_shiftram
    GENERIC MAP (
      g_data_w                           => 2*c_acc_data_w,
      g_nof_words                        => pow2(ceil_log2(g_nof_accumulators)),
      g_output_invalid_during_shift_incr => TRUE,
      g_fixed_shift                      => TRUE
    )
    PORT MAP (
      rst            => rst,
      clk            => clk,
      
      data_in        => common_shiftram_snk_in_arr(i).data(2*c_acc_data_w-1 DOWNTO 0),
      data_in_val    => common_shiftram_snk_in_arr(i).valid,
      data_in_shift  => c_common_shiftram_shift_in,
   
      data_out       => common_shiftram_src_out_arr(i).data(2*c_acc_data_w-1 DOWNTO 0),
      data_out_val   => common_shiftram_src_out_arr(i).valid,
      data_out_shift => common_shiftram_data_out_shift_arr(i)
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Output g_integration_period per stream once per integration period
  -- . The first g_nof_accumulators words are output. At the same time, the
  --   accumulators are reset (zeros) at the adder inputs.
  -- . Make sure the shiftram output is valid too so we don't output a block
  --   of zeros initially.
  -----------------------------------------------------------------------------
  gen_src_out_arr : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    nxt_src_out_arr(i).re(c_acc_data_w-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(2*c_acc_data_w-1 DOWNTO c_acc_data_w);
    nxt_src_out_arr(i).im(c_acc_data_w-1 DOWNTO 0) <= common_shiftram_src_out_arr(i).data(  c_acc_data_w-1 DOWNTO 0);
    
    nxt_src_out_arr(i).valid <= '1' WHEN TO_UINT(acc_cnt)<g_nof_accumulators   AND common_shiftram_src_out_arr(0).valid='1' ELSE '0';
    nxt_src_out_arr(i).sop   <= '1' WHEN TO_UINT(acc_cnt)=0                    AND common_shiftram_src_out_arr(0).valid='1' ELSE '0';
    nxt_src_out_arr(i).eop   <= '1' WHEN TO_UINT(acc_cnt)=g_nof_accumulators-1 AND common_shiftram_src_out_arr(0).valid='1' ELSE '0';
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Registers
  -----------------------------------------------------------------------------
  p_clk : PROCESS (clk, rst)
  BEGIN
    IF rst='1' THEN
      acc_cnt                  <= (OTHERS=>'0');
      corr_adder_snk_in_2arr_2 <= (OTHERS=>(OTHERS=>c_dp_sosi_rst));
      src_out_arr              <= (OTHERS=>c_dp_sosi_rst);
    ELSIF rising_edge(clk) THEN
      acc_cnt                  <= nxt_acc_cnt; 
      corr_adder_snk_in_2arr_2 <= nxt_corr_adder_snk_in_2arr_2;
      src_out_arr              <= nxt_src_out_arr;
    END IF;
  END PROCESS;

END str;
