--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Purpose:
-- . Buffer the visibilities during folding 
-- Description:
-- . Consists of an array of FIFOs that are read out periodically.
-- . The output data rate is such that all output streams can be folded into
--   one stream (valid cycles/stream never occur in parallel)
-- . In addition, the FIFOs are read out such that the folder outputs channel
--   blocks of visibilities.
-- . Finally, the output stream is toggled to produce a constant flow
--   instead of a short burst (to the extent that it is allowed by the 
--   integration time - a short integration time leaves little headroom).

ENTITY corr_visibility_buffer IS
  GENERIC (
    g_nof_inputs         : NATURAL; -- Number of input streams
    g_buffer_depth       : NATURAL; -- Number of words to buffer
    g_data_w             : NATURAL; -- Complex input data width
    g_nof_pre_mult_folds : NATURAL; -- Nof times the data has been folded
    g_integration_period : NATURAL
   ); 
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0)
  );
END corr_visibility_buffer;

ARCHITECTURE str OF corr_visibility_buffer IS

  -- The output block size required so the folder outputs all visibilities per channel
  CONSTANT c_output_block_len     : NATURAL := pow2(g_nof_pre_mult_folds);

  -- The minimum period to wait so the folder does not miss data
  CONSTANT c_minimum_block_period : NATURAL := g_nof_inputs*c_output_block_len;

  -- The channel blocks on the folder output will be c_minimum_block_period long.
  -- Wait the maximum time before outputting the next one to yield a constant data rate.
  -- Note: insertion of a gap block is only possible if 
  --       g_integration_period/2 >= c_minimum_block_period.
  CONSTANT c_block_valid_ratio    : NATURAL := largest(1, g_integration_period/c_minimum_block_period);

  SIGNAL dp_fifo_sc_src_out_arr   : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL dp_fifo_sc_src_in_arr    : t_dp_siso_arr(g_nof_inputs-1 DOWNTO 0);

BEGIN

  ------------------------------------------------------------------------------
  -- Buffer the visibilities before the folding stage.
  -- . All accumulators output blocks of visibilities at the same time. We need
  --   to buffer the visibilities to source them to the folder at the right
  --   pace.
  ------------------------------------------------------------------------------
  gen_dp_fifo_sc : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_dp_fifo_sc : ENTITY dp_lib.dp_fifo_sc
    GENERIC MAP (
      g_data_w         => 2*g_data_w,
      g_use_ctrl       => FALSE,
      g_use_complex    => TRUE,
      g_fifo_size      => g_buffer_depth,
      g_fifo_af_margin => 0
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,

      wr_ful      => OPEN,
      usedw       => OPEN,
      rd_emp      => OPEN,

      snk_out     => OPEN,
      snk_in      => snk_in_arr(i),
      src_in      => dp_fifo_sc_src_in_arr(i),
      src_out     => dp_fifo_sc_src_out_arr(i)
    );
  END GENERATE;

  ------------------------------------------------------------------------------
  -- Read out the FIFOs such that dp_fifo_sc_src_out_arr only carries one valid 
  -- word per cycle. This allows folding (multiplexing) into one stream without 
  -- losing data (example for 16 streams with c_output_block_len=1):            
  --                                                                            
  --      |<---c_minimum_block_period---->|                                     
  --   0) [0]                             [1]   ..                              
  --   1) | [0]                           . [1]                                 
  --   2) |   [0]                         .   [1]                               
  --   3) |     [0]                       .     [1]                             
  --   4) |       [0]                     .       [1]                           
  --   5) |         [0]                   .         [1]                         
  --   6) |           [0]                 .           [1]                       
  --   7) |             [0]               .             [1]                     
  --   8) |               [0]             .               [1]                   
  --   9) |                 [0]           .                 [1]                 
  --  10) |                   [0]         .                   [1]               
  --  11) |                     [0]       .                     [1]             
  --  12) |                       [0]     .                       [1]           
  --  13) |                         [0]   .                         [1]         
  --  14) |                           [0] .                           [1]       
  --  15) |<---g_init_valid_delay(15)-->[0]                             [1]     
  ------------------------------------------------------------------------------
  gen_dp_src_out_timer : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_dp_src_out_timer : ENTITY dp_lib.dp_src_out_timer
    GENERIC MAP (
      g_init_valid_delay => i*c_output_block_len, --relative to dp_fifo_sc_src_out_arr(i).valid
      g_block_period     => c_block_valid_ratio*c_minimum_block_period*c_output_block_len,
      g_block_len        => c_output_block_len
    )
    PORT MAP (
      rst                  => rst,
      clk                  => clk,

      init_valid_delay_ref => snk_in_arr(i).valid,

      snk_in               => dp_fifo_sc_src_out_arr(i),
      snk_out              => dp_fifo_sc_src_in_arr(i)
    );
  END GENERATE;

  src_out_arr <= dp_fifo_sc_src_out_arr;

END str;
