-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Testbench for tech_10gbase_r.
-- Description:
-- . The tb is not self checking.
-- . The tb is used to investigate what clocks the 10gbase_r phy uses.
-- . The link fault is modelled on channel 0 only. Channel >= 1 is connected
--   from the start and remains connected.
-- Usage:
--   > as 10
--   > run -a
-- Observations:
-- . With the link_fault it appears that tx_ready and rx_ready do not depend
--   on whether the link is connected or not.
--   A link fault does cause XGMII control to become 0x11 and data to become
--   0x0100009C_0100009C. When the link is connected or reconnected then the
--   XGMII control depends on xgmii_tx_dc_arr.

LIBRARY IEEE, technology_lib, tech_pll_lib, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_interface_layers_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_pll_lib.tech_pll_component_pkg.ALL;


ENTITY tb_tech_10gbase_r IS
  -- Test bench control parameters
  GENERIC (
    g_technology          : NATURAL := c_tech_select_default
  );
END tb_tech_10gbase_r;


ARCHITECTURE tb OF tb_tech_10gbase_r IS

  CONSTANT c_sim               : BOOLEAN:= TRUE;
  CONSTANT c_sim_level         : NATURAL := 0;     -- 0 = use IP; 1 = use fast serdes model (not useful here, because no proper xgmii_tx_dc_arr stimuli)
  CONSTANT phy_loopback_delay  : TIME :=  sel_a_b(c_sim_level=0, 1 ns, 0 ns);
  CONSTANT c_nof_channels      : NATURAL := 2;
  
  SIGNAL tb_end             : STD_LOGIC := '0';
  SIGNAL tr_ref_clk_644     : STD_LOGIC := '0';
  SIGNAL clk_156            : STD_LOGIC;
  SIGNAL rst_156            : STD_LOGIC;

  -- XGMII interface
  SIGNAL xgmii_tx_ready_arr : STD_LOGIC_VECTOR(c_nof_channels-1 DOWNTO 0);
  SIGNAL xgmii_rx_ready_arr : STD_LOGIC_VECTOR(c_nof_channels-1 DOWNTO 0);
  SIGNAL xgmii_tx_dc_arr    : t_xgmii_dc_arr(c_nof_channels-1 DOWNTO 0) := (OTHERS=>(OTHERS=>'0'));  -- '0', '1'
  SIGNAL xgmii_rx_dc_arr    : t_xgmii_dc_arr(c_nof_channels-1 DOWNTO 0);

  -- PHY serial interface
  --SIGNAL link_fault         : STD_LOGIC := '1';   -- model initial link fault
  SIGNAL link_fault         : STD_LOGIC := '0';
  SIGNAL tx_serial_arr      : STD_LOGIC_VECTOR(c_nof_channels-1 DOWNTO 0);
  SIGNAL tx_serial_arr_dly  : STD_LOGIC_VECTOR(c_nof_channels-1 DOWNTO 0);
  SIGNAL rx_serial_arr      : STD_LOGIC_VECTOR(c_nof_channels-1 DOWNTO 0);
  
BEGIN

  tr_ref_clk_644 <= NOT tr_ref_clk_644 AFTER tech_pll_clk_644_period/2;

  p_stimuli : PROCESS
  BEGIN
    WAIT FOR 50 us;
    -- Model a link fault to verify Rx recovery
    link_fault <= '1';
    WAIT FOR 10 us;
    link_fault <= '0';
    WAIT FOR 10 us;
    tb_end <= '1';
    REPORT "Tb simulation finished." SEVERITY FAILURE;
    WAIT;
  END PROCESS;

  pll : ENTITY tech_pll_lib.tech_pll_xgmii_mac_clocks
  GENERIC MAP (
    g_technology => g_technology
  )
  PORT MAP (
    refclk_644 => tr_ref_clk_644,
    clk_156    => clk_156,
    clk_312    => OPEN,
    rst_156    => rst_156,
    rst_312    => OPEN
  );

  dut : ENTITY work.tech_10gbase_r
  GENERIC MAP (
    g_technology       => g_technology,
    g_sim              => c_sim,
    g_sim_level        => c_sim_level,
    g_nof_channels     => c_nof_channels
  )
  PORT MAP (
    -- Transceiver ATX PLL reference clock
    tr_ref_clk_644     => tr_ref_clk_644,
    
    -- XGMII clocks
    clk_156            => clk_156,
    rst_156            => rst_156,

    -- XGMII interface
    xgmii_tx_ready_arr => xgmii_tx_ready_arr,
    xgmii_rx_ready_arr => xgmii_rx_ready_arr,
    xgmii_tx_dc_arr    => xgmii_tx_dc_arr,
    xgmii_rx_dc_arr    => xgmii_rx_dc_arr,

    -- PHY serial IO
    tx_serial_arr      => tx_serial_arr,
    rx_serial_arr      => rx_serial_arr
  );
  
  -- PHY loopback
  tx_serial_arr_dly <= TRANSPORT tx_serial_arr AFTER phy_loopback_delay;
  
  p_link : PROCESS(tx_serial_arr_dly, link_fault)
  BEGIN
    rx_serial_arr <= tx_serial_arr_dly;
    IF link_fault='1' THEN
      rx_serial_arr(0) <= '0';  -- model link fault only for channel 0
    END IF;
  END PROCESS;
  
END tb;
